#!/usr/bin/env python
# encoding: utf-8
import subprocess
import might.might_log as mlog
from might.miscellaneous import check_for_dependency, file_check, ramdisk_handler


def kraken2_precheck(kraken2_database=None, force=False, kraken2_directory=None, kraken2_report=None, r1_fastq=None,
                     r2_fastq=None, colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None):
    """
    sets all of the file/directory path attributes associated with a kraken2 analysis. If the kraken2_database value
    is NOT provided then the analysis will be aborted. If the ramdisk value is provided, set the attributes
    necessary for running kraken2 with --memory-mapping
    """

    if not check_for_dependency('kraken2'):
        mlog.log('Could not find kraken2 on your PATH, so this analysis is being aborted',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    if not file_check(r1_fastq) or not file_check(r2_fastq):
        mlog.log('Could not find both read files, so this analysis is being aborted',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    if not kraken2_database:
        mlog.log('The path to the kraken2 database was not provided, so this analysis is being aborted',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    # verify that the kraken2_database path provided is valid
    if not file_check(kraken2_database):
        mlog.log('The path to the kraken2 database appears to be invalid, so this analysis is being aborted',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    # check to see if the kraken2_directory exists already. If so, check to see if the output kraken2_report is already
    # present. If not, make the directory
    if file_check(kraken2_directory):
        if file_check(kraken2_report):
            if force:
                mlog.log('Kraken2 output for this sample was already present, but --force was invoked so we will '
                         'overwrite the existing results',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                subprocess.run(['rm', str(kraken2_report)])
            else:
                mlog.log('Kraken2 output for this sample was already present, so this analysis is being aborted',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                return False
    else:
        mlog.log('Preparing the kraken2 output directory...\n',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        if not file_check(str(kraken2_directory)):
            subprocess.run(['mkdir', str(kraken2_directory)])

    return True


def run_kraken2(kraken2_database=None, kraken2_report=None, r1_fastq=None, r2_fastq=None, ramdisk=None, threads='1',
                colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None):
    """
    run kraken2
    """

    if ramdisk_handler(mount=True,
                       ramdisk_directory=ramdisk,
                       directory_to_load=kraken2_database,
                       colors=colors,
                       stdout_verbosity_level=stdout_verbosity_level,
                       log_file_verbosity_level=log_file_verbosity_level,
                       log_file=log_file):

        # run kraken2 command with ramdisk utility
        kraken2_cmd = 'kraken2 --db ' + str(ramdisk / kraken2_database.name) + ' --memory-mapping --use-names --report '\
                      + str(kraken2_report) + \
                      ' --threads ' + threads + ' --paired  ' + str(r1_fastq) + ' ' + str(r2_fastq)
        mlog.log('Command: ' + mlog.bold(kraken2_cmd),
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        mlog.log('Now running kraken2...\n',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        kraken2_process = subprocess.run(['kraken2',
                                          '--db', str(ramdisk / kraken2_database.name),
                                          '--memory-mapping',
                                          '--use-names',
                                          '--report', str(kraken2_report),
                                          '--threads', threads,
                                          '--paired', str(r1_fastq), str(r2_fastq)],
                                         stdout=subprocess.DEVNULL,
                                         stderr=subprocess.PIPE,
                                         universal_newlines=True)
        mlog.log(str(kraken2_process.stderr),
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
    else:
        # run kraken2 command without ramdisk utility

        kraken2_cmd = 'kraken2 --db  ' + str(kraken2_database) + ' --use-names --report  ' + str(kraken2_report) + \
                      ' --threads 1 --paired  ' + str(r1_fastq) + ' ' + str(r2_fastq)
        mlog.log('Command: ' + mlog.bold(kraken2_cmd),
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        mlog.log('Now running kraken2...\n',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        kraken2_process = subprocess.run(['kraken2',
                                          '--db', str(kraken2_database),
                                          '--use-names',
                                          '--report', str(kraken2_report),
                                          '--threads', threads,
                                          '--paired', str(r1_fastq), str(r2_fastq)],
                                         stdout=subprocess.DEVNULL,
                                         stderr=subprocess.PIPE,
                                         universal_newlines=True)
        mlog.log(str(kraken2_process.stderr),
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # validate that an output file was generated, and update sample.kraken2_status accordingly
    if not file_check(kraken2_report):
        mlog.log('Kraken2 failed to produce a report. See log file for details',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    mlog.log('Kraken2 analysis is complete',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)
    return




#!/usr/bin/env python
# encoding: utf-8
import subprocess
from pathlib import Path
from Bio import SeqIO
import might.might_log as mlog
from might.miscellaneous import check_for_dependency, file_check


def assembly_precheck(r1_fastq=None, r2_fastq=None, raw_long_reads_file=None, individual_assembly_directory=None,
                      assembly=None, force=False, colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1,
                      log_file=None):
    """
    for assembly to run, must have:
        1) unicycler on PATH
        2) r1_fastq and r2_fastq detected
        3) output directory does NOT already exist
    """

    # validate that the bbduk.sh adapter trimming program is on the PATH
    if not check_for_dependency('bbduk.sh'):
        mlog.log('Could not find bbduk on your PATH, so this analysis is being aborted',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False
    else:
        mlog.log('utility located: bbduk',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # validate that the unicycler assembly program is on the PATH
    if not check_for_dependency('unicycler'):
        mlog.log('Could not find Unicycler on your PATH, so this analysis is being aborted',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False
    else:
        mlog.log('utility located: unicycler',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # validate that the illumina paired end read files are present
    if not file_check(r1_fastq) or not file_check(r2_fastq):
        mlog.log('Could not find both read files, so this analysis is being aborted',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False
    else:
        mlog.log('R1 and R2 read files located',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # validate that the raw_long_reads file exists
    if not file_check(raw_long_reads_file):
        mlog.log('The long reads file could not be located, so the long reads will not be used!',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        raw_long_reads_file = None
    else:
        mlog.log('raw long reads files located',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # check to see if the individual_assembly_directory exists already. If so, check to see if the output assembly.fasta
    # is already present.
    if file_check(individual_assembly_directory):
        if file_check(assembly):
            if force:
                mlog.log('Unicycler output for this sample was already present, but --force was invoked so we will '
                         'overwrite the existing results',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                subprocess.run(['rm', str(individual_assembly_directory)])
            else:
                mlog.log('Unicycler output for this sample was already present, so this analysis is being aborted\n',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                return False

    mlog.log('Assembly precheck passed: sample ready for adapter trimming and assembly!',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    # return the assembly type. If short reads ONLY are present, return "short". If BOTH short and long reads are
    # present, return "long"

    if raw_long_reads_file:
        return 'long'
    else:
        return 'short'


def adapter_trimming(r1_fastq=None, r2_fastq=None, r1_fastq_trimmed=None, r2_fastq_trimmed=None, adapter_file=None,
                     force=False, colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None):
    """
    perform adapter trimming on the raw fastq files
    """

    mlog.log_section_header('Adapter Trimming',
                            colors=colors,
                            stdout_verbosity_level=stdout_verbosity_level,
                            log_file_verbosity_level=log_file_verbosity_level,
                            log_file=log_file)
    mlog.log_explanation('Perform adapter (and light quality) trimming of Illumina read files using bbduk.',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)

    # validate the adapter_fa_file
    if not file_check(adapter_file):
        mlog.log('cannot locate the adapter.fa file for bbduk trimming, aborting assembly',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    # check for preexisting adapter trimmed files
    if file_check(r1_fastq_trimmed) and file_check(r2_fastq_trimmed):
        if force:
            subprocess.run(['rm', str(r1_fastq_trimmed)])
            subprocess.run(['rm', str(r2_fastq_trimmed)])
        else:
            mlog.log('trimmed read files located, so skipping adapter trimming with bbduk',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            return True

    # if only one of the two trimmed read files is present, remove it and redo the trimming
    if file_check(r1_fastq_trimmed) and not file_check(r2_fastq_trimmed):
        subprocess.run(['rm', str(r1_fastq_trimmed)])
    elif not file_check(r1_fastq_trimmed) and file_check(r2_fastq_trimmed):
        subprocess.run(['rm', str(r2_fastq_trimmed)])

    # prepare the adapter trimming cmd as a string
    adapter_trimming_cmd = 'bbduk.sh in1=' + str(r1_fastq) + ' in2=' + str(r2_fastq) + ' out1=' + str(r1_fastq_trimmed) \
                           + " out2=" + str(r2_fastq_trimmed) + " ref= " + str(adapter_file) + 'ktrim=r k=23' \
                            ' mink=11 hdist=1 tpe tbo qtrim=r trimq=8'

    mlog.log('Command: ' + mlog.bold(adapter_trimming_cmd) + '\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    # run the adapter trimming command
    bbduk_proc = subprocess.run(['bbduk.sh',
                                 'in1=' + str(r1_fastq),
                                 'in2=' + str(r2_fastq),
                                 'out1=' + str(r1_fastq_trimmed),
                                 'out2=' + str(r2_fastq_trimmed),
                                 'ref=' + str(adapter_file),
                                 'ktrim=r',
                                 'k=23',
                                 'mink=11',
                                 'hdist=1',
                                 'tpe',
                                 'tbo',
                                 'qtrim=r',
                                 'trimq=8'],
                                stderr=subprocess.PIPE,
                                universal_newlines=True)

    bbduk_stderr = bbduk_proc.stderr.split('Version')[1].splitlines()
    mlog.log('Version ' + bbduk_stderr[0],
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)
    for line in bbduk_stderr[1:]:
        mlog.log(line,
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # verify the output of the adapter trimming cmd
    if not file_check(r1_fastq_trimmed) or not file_check(r2_fastq_trimmed):
        mlog.log('bbduk failed to output adapter trimmed read file(s), aborting assembly',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    return True


def run_unicycler_assembly(r1_fastq_trimmed=None, r2_fastq_trimmed=None, long_read_file=None, mode='short',
                           individual_assembly_directory=None, threads='1', colors=1, stdout_verbosity_level=1,
                           log_file_verbosity_level=1, log_file=None):
    """
    run assembly using Unicycler.
    """

    # prepare the Unicycler command
    if mode == 'short':
        unicycler_command = 'unicycler --short1 ' + str(r1_fastq_trimmed) + ' --short2 ' + str(r2_fastq_trimmed) + \
                            ' --out ' + str(individual_assembly_directory) + ' --verbosity 1 '
    elif mode == 'long':
        unicycler_command = 'unicycler --short1 ' + str(r1_fastq_trimmed) + ' --short2 ' + str(r2_fastq_trimmed) + \
                            '--long' + str(long_read_file) + ' --out ' + str(individual_assembly_directory) +\
                            ' --verbosity 1 '

    # run Unicycler command
    if mode == 'short':
        subprocess.run(['unicycler',
                        '--short1', str(r1_fastq_trimmed),
                        '--short2', str(r2_fastq_trimmed),
                        '--out', str(individual_assembly_directory),
                        '--threads', threads,
                        '--verbosity', str(stdout_verbosity_level),
                        '--keep', '3'])
    elif mode == 'long':
        subprocess.run(['unicycler',
                        '--short1', str(r1_fastq_trimmed),
                        '--short2', str(r2_fastq_trimmed),
                        '--long', str(long_read_file),
                        '--out', str(individual_assembly_directory),
                        '--threads', threads,
                        '--verbosity', str(stdout_verbosity_level),
                        '--keep', '3'])


def assembly_file_formatting(assembly_file=None, sample_name=None, colors=1, stdout_verbosity_level=1,
                             log_file_verbosity_level=1, log_file=None):
    """
    The default unicycler assembly output contains a file "assembly.fasta" which is a multi-fasta file containing all
    of the assembled contigs. By default, each of the contig id/description lines is of the format:
        >[contig #] length=[contig length] depth=[relative depth]x
    We want to update the [contig #] field to fit into later aspects of our pipeline, so we will append the sample_name
    and contig_ with leading zeros to the output.

    We will write the reformatted contig name/descriptions to a new file in the same directory called
    [sample_name]_assembly.fasta

    :param assembly_file: POSIX path to the assembly file (fasta file output of assembler) to be reformatted
    :param sample_name: Name of the sample that was assembled
    :return: Posix path of the reformatted file if renaming was successful, the original path if other
    """

    # validate that there is a file at the assembly_file location, and if so prepare the reformatted file path
    if not assembly_file:
        mlog.log('No assembly file was provided!',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return assembly_file

    if not assembly_file.exists():
        mlog.log('Could not locate the assembly file at the specified path: ' + assembly_file,
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return assembly_file

    reformatted_assembly_file = assembly_file.parent / (sample_name + '_assembly.fasta')

    if reformatted_assembly_file.exists():
        mlog.log('The reformatted assembly file was located, so no reformatting will be performed\n',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return reformatted_assembly_file

    # read in all of the contig information from assembly_file using SeqIO
    mlog.log('Reformatting the assembly file now...',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)
    with open(assembly_file, 'r') as original, open(reformatted_assembly_file, 'w') as reformatted:
        contigs = SeqIO.parse(assembly_file, 'fasta')
        for contig in contigs:
            contig_id_element_list = contig.description.split(' ')
            contig_id_element_list[0] = sample_name + '_contig' + \
                                                 str(contig_id_element_list[0]).zfill(4)
            contig.id = contig_id_element_list[0]
            contig.description = ' '.join(contig_id_element_list[1:])
            SeqIO.write(contig, reformatted, 'fasta')

    mlog.log('Assembly file reformatting complete!',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    return reformatted_assembly_file


def filter_long_reads(r1_fastq=None, r2_fastq=None, raw_long_reads=None, processed_long_reads=None, colors=1,
                      stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None, force=False):
    """
    filtlong will be run with the following settings:
        min_length=10,000
        target_bases=500,000,000
        trim
        split=500
    """

    mlog.log_section_header('Long Read Data Trimming',
                            colors=colors,
                            stdout_verbosity_level=stdout_verbosity_level,
                            log_file_verbosity_level=log_file_verbosity_level,
                            log_file=log_file)
    mlog.log_explanation('Use filtlong to perform quality and length filtering of the provided long read sequencing '
                         'data',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)

    # check to see that filtlong is available on the users PATH
    if not check_for_dependency('filtlong'):
        mlog.log('Could not find filtlong on your PATH, so the long reads will not be trimmed!',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False
    else:
        mlog.log('utility located: filtlong',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # validate that the input requirements are met
    if not r1_fastq or not r2_fastq or not raw_long_reads:
        mlog.log('One or more of the required input files could not be located, so the long reads will not be trimmed!',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    # check for preexisting adapter trimmed files
    if file_check(processed_long_reads):
        if force:
            subprocess.run(['rm', str(processed_long_reads)])
        else:
            mlog.log('trimmed long read file located, so skipping trimming with filtlong',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            return True

    # log the filtlong command line input
    filtlong_command = "filtlong -1 " + str(r1_fastq)+ " -2 " + r2_fastq + " --min_length 10000 --target_bases " \
                       "500000000 --trim --split 500 " + raw_long_reads + " > " + str(processed_long_reads)

    mlog.log('Command: ' + mlog.bold(filtlong_command) + '\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    filtlong_proc = subprocess.run(['filtlong',
                                    '-1', str(r1_fastq),
                                    '-2', str(r2_fastq),
                                    '--min_length', '10000',
                                    '--target_bases', '500000000',
                                    '--trim',
                                    '--split', '500',
                                    str(raw_long_reads),
                                    '>', str(processed_long_reads)],
                                   shell=True,
                                   stderr=subprocess.PIPE,
                                   universal_newlines=True)

    return True


def assembly_QC(r1_fastq_trimmed=None, r2_fastq_trimmed=None, long_read_file=None, mode='short', output_directory=None,
                assembly=None, qc_directory=None, threads='1', colors=1, stdout_verbosity_level=1,
                log_file_verbosity_level=1, log_file=None):

    mlog.log_section_header('Assembly QC',
                            colors=colors,
                            stdout_verbosity_level=stdout_verbosity_level,
                            log_file_verbosity_level=log_file_verbosity_level,
                            log_file=log_file)
    mlog.log_explanation('Use QUAST to perform QC analysis of the assembly generated by unicycler',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)

    # check to see that QUAST is available on the users PATH
    if not check_for_dependency('quast'):
        mlog.log('Could not find quast on your PATH, so assembly QC can not be performed!',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False
    else:
        mlog.log('utility located: quast',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # ensure that all of the necessary file paths have been provided and are valid
    if not r1_fastq_trimmed or not r2_fastq_trimmed or not assembly:
        mlog.log('Not all of the requisite files were provided for assembly QC!',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    if not file_check(r1_fastq_trimmed) or not file_check(r2_fastq_trimmed) or not file_check(assembly):
        mlog.log('One or more of the specified input files could not be located at the provided location!',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return False

    # if the mode is set to 'long', verify that the long read file path has been provided and is valid
    if mode == 'long':
        if not long_read_file:
            mlog.log('The path to the long read file was not provided, but the assembly QC request type was set to long'
                     '. Since the short reads and the assembly file were provided, we will use these to perform the '
                     'assembly QC.',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            mode = 'short'

        if not file_check(long_read_file):
            mlog.log('The path to the long read file provided is invalid, but the assembly QC request type was set to '
                     'long. Since the short reads and the assembly file were provided, we will use these to perform the'
                     ' assembly QC.',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            mode = 'short'

    # check to see if a qc_directory was specified. If so we will need to make sure that it doesn't already exist. If it
    # already exists or the user didn't specify a directory, we will default to a directory labelled "QC" in the output
    # directory if the output directory is specified or the parent of the assembly file is not.
    if qc_directory:
        if Path(qc_directory).expanduser().resolve().exists():
            mlog.log('The path provided for the QC directory already exists! Rather than overwrite it, we will '
                     'substitute the default QC directory.',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            qc_directory = None

    if not qc_directory:
        if output_directory:
            if file_check(output_directory):
                qc_directory = Path(output_directory).expanduser().resolve() / 'QC'
        else:
            qc_directory = Path(assembly).expanduser().resolve().parent / 'QC'

    # Log the quast command based on the mode (i.e. whether or not to include the long read data)
    if mode == 'short':
        quast_cmd = 'quast --min-contig 200 --threads ' + str(threads) + ' --pe1 ' + str(r1_fastq_trimmed) + ' --pe2 ' \
                    + str(r2_fastq_trimmed) + ' -o ' + str(qc_directory) + ' ' + str(assembly)
    elif mode == 'long':
        quast_cmd = 'quast --min-contig 200 --threads ' + str(threads) + ' --pe1 ' + str(r1_fastq_trimmed) + ' --pe2 ' \
                    + str(r2_fastq_trimmed) + ' --pacbio ' + str(long_read_file) + ' -o ' + str(qc_directory) + ' ' + \
                    str(assembly)

    mlog.log('Command: ' + mlog.bold(quast_cmd) + '\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    # run the quast command
    subprocess.run(quast_cmd, shell=True, stderr=subprocess.STDOUT, stdout=subprocess.DEVNULL)





#!/usr/bin/env python
# encoding: utf-8
import fnmatch
import os
import subprocess
from pathlib import Path
import might.might_log as mlog
from might.amr_methods import amr_precheck, ariba_run, amrfinderplus_run, summarize_amr_finder_output
from might.assembly_methods import assembly_precheck, adapter_trimming, run_unicycler_assembly, assembly_file_formatting, \
    filter_long_reads
from might.database_methods import database_check
from might.kraken_methods import kraken2_precheck, run_kraken2
from might.miscellaneous import parse_arguments, compression_handler, file_check, ascii_art


def premain():
    ascii_art(logo='might')
    args = parse_arguments(might=True)
    # use args to instantiate the IndividualSequencingSample object
    sample = IndividualSequencingSample(args)
    main(sample)


def main(sample):

    kraken2(sample)

    assembly(sample)

    amr(sample)

    cleanup(sample)


class IndividualSequencingSample:
    """
    object used to track the progress of an individual sample/isolate through the analysis pipeline. Primary usage is
    to determine if the input/output requirements of a given analysis are met before launching
    """

    def __init__(self, args):

        # self.name is the name of the sequencing sample. Provided at the command line, it should be the first field in
        # the fastq/fasta file followed by an '_'
        self.name = args.sample_name

        # self.output_directory is the parent directory which will store all of the output for this sample, as well as
        # any other samples being processed concurrently with AllMight.py. If it doesn't already exist, we will make it
        # here
        self.output_directory = Path(args.output).expanduser().resolve()
        if not self.output_directory.exists():
            subprocess.run(['mkdir', str(self.output_directory)])

        # self.logs_folder is output subdirectory where all of the might log files are stored
        self.logs_folder = self.output_directory / 'logs'
        if not self.logs_folder.exists():
            subprocess.run(['mkdir', str(self.logs_folder)])

        # self.log is the path to the log file for this sample
        self.log_file = self.logs_folder / (self.name + '.might.log')

        # if the verbosity levels are specified
        self.stdout_verbosity = int(args.verbosity)

        self.log_file_verbosity = 4

        # ALT: Use the this object to track the log object attributes, and use the static methods
        mlog.initialize_logger_attributes(self,
                                          log_filename=self.log_file,
                                          stdout_verbosity_level=self.stdout_verbosity,
                                          log_file_verbosity_level=self.log_file_verbosity)

        # self.reads_directory is the parent directory for all read files that are GENERATED during an analysis. These
        # include raw_reads if bcl2fastq2 was run via AllMight.py as well as processed reads if read trimming was
        # performed as part of the assembly analysis
        self.reads_directory = self.output_directory / 'reads'
        if not self.reads_directory.exists():
            subprocess.run(['mkdir', str(self.reads_directory)])

        # self.raw_reads_directory is set at the command line, and could refer to a directory within the output/reads
        # directory (in the case where AllMight.py generated fastqs via bcl2fastq2) or a user specified directory where
        # the read files for the current sample are located
        if args.fastq:
            self.raw_reads_directory = Path(args.fastq).expanduser().resolve()
        else:
            self.raw_reads_directory = self.reads_directory / 'raw_reads'

        # self.processed_reads_directory is a subdirectory of the reads directory and will contain any trimmed read
        # files that are generated if --assembly is specified
        self.processed_reads_directory = self.reads_directory / 'processed_reads'

        # self.r1_fastq and self.r2_fastq refer to the Illumina paired end R1 and R2 read files, respectively. They are
        # initialized as None type and then automatically assigned the file paths using scan_for_fastq(), assuming the
        # files exist
        self.r1_fastq = None
        self.r2_fastq = None
        if args.fastq:
            self.scan_for_fastq()

        # self.raw_long_reads_file refers to the user specified long reads file in fastq format
        if args.long_reads:
            self.raw_long_reads_file = args.long_reads

        # self.assembly is used to store the path to the assembled contigs file. It is initialized as None type and then
        # automatically assigned the file paths using scan_for_fasta(), assuming the file exists
        self.assembly_file = None
        if args.fasta:
            self.scan_for_assembly(args.fasta)

        # self.force is set at the command line. If set to true, it will overwrite all existing output it comes across
        # for this sample. Use with caution!
        self.force = args.force

        # self.ramdisk will store the path to the prepared ramdisk if provided at the command line, or None if it was
        # not specified.
        if args.ramdisk:
            self.ramdisk = Path(args.ramdisk).expanduser().resolve()
        else:
            self.ramdisk = None

        # self.cores is the number of cores, stored as a string, that are available for this sample's analysis
        self.cores = str(args.cores)

        # kraken2 analysis specific path attributes. They are only set to non-None values if Kraken2 analysis is
        # requested. kraken2 database presence is validated during argument parsing
        self.kraken2 = args.kraken2
        if self.kraken2:

            self.kraken2_database = Path(args.kraken2_database).expanduser().resolve()

            # self.kraken2_directory is the parent directory for all kraken2 output
            self.kraken2_directory = self.output_directory / 'kraken2'

            # self.kraken2_report is the path to the output of a kraken2 analysis run
            self.kraken2_report = self.kraken2_directory / (self.name + '_kraken2_report.txt')

        # assembly specific file paths. Like the Kraken2 attributes these file path attributes are only instantiated if
        # assembly is requested
        self.assembly = args.assembly
        if args.assembly:

            # self.assembly_directory is the parent directory for all assembly output
            self.assembly_directory = self.output_directory / 'assembly'

            # self.individual_assembly_directory is the parent directory for all files and directories generated during
            # assembly pertaining to this sample
            self.individual_assembly_directory = self.assembly_directory / self.name

            # self.adapter_file
            self.adapter_file = Path(args.adapter_file).expanduser().resolve()
            self.r1_fastq_trimmed = self.processed_reads_directory / (self.name + '_r1_trimmed.fastq')
            self.r2_fastq_trimmed = self.processed_reads_directory / (self.name + '_r2_trimmed.fastq')

        # amr specific file paths, instantiated only if amr is analysis is requested
        self.amr = args.amr
        if args.amr:

            # self.amr_directory is the parent directory for all amr analysis. If it doesn't exist, generate it
            self.amr_directory = self.output_directory / 'amr'
            if not self.amr_directory.exists():
                subprocess.run(['mkdir', str(self.amr_directory)])

            # self.individual_amr_directory is the parent directory for all files and directories generated during
            # amr analysis of this sample
            self.individual_amr_directory = self.amr_directory / self.name

            # self.ariba_output_directory stores all of the output files and directories generated by ARIBA if amr
            # analysis is performed using read data
            self.ariba_output_directory = self.individual_amr_directory / 'ARIBA_output'
            self.ariba_assembled_genes = self.ariba_output_directory / 'assembled_genes.fa.gz'
            self.ariba_assemblies = self.ariba_output_directory / 'assemblies.fa.gz'

            # file paths to all of the amr analysis output files
            self.amrfinder_output_from_ariba = self.individual_amr_directory / 'amrfinder_output_from_ariba.txt'
            self.amrfinder_output_from_contigs = self.individual_amr_directory / 'amrfinder_output_from_contigs.txt'
            self.amr_output_file = self.individual_amr_directory / 'andale_output_raw.csv'

        # mlst specific file paths
        if args.mlst:

            self.mlst = args.mlst

            self.mlst_directory = self.output_directory / 'mlst'
            self.mlst_assignment_file = ""

        # plasmidfinder specific file paths
        if args.plasmidfinder:

            self.plasmidfinder = args.plasmidfinder

            self.plasmidfinder_directory = self.output_directory / 'plasmidfinder'
            self.plasmidfinder_results_file = ""

        self.pbs_script_directory = self.output_directory / 'pbs_scripts'

        self.update = args.update

    def scan_for_fastq(self, directory=None):
        """
        scans the specified directory (default is output/reads/raw_reads/) for fastq files that match the name of the
        sample, then assigns their POSIX paths to the self.r1_fastq and self.r2_fastq object attributes. Finally, method
        returns True if both the r1 and r2 files were located, False if they were not
        """
        if directory:
            directory = Path(directory).expanduser().resolve()
        else:
            directory = self.raw_reads_directory

        for file in os.scandir(directory):
            if file.name.startswith(self.name+"_") and fnmatch.fnmatch(file, '*R1*fastq*'):
                self.r1_fastq = Path(file).expanduser().resolve()
            elif file.name.startswith(self.name+"_") and fnmatch.fnmatch(file, '*R2*fastq*'):
                self.r2_fastq = Path(file).expanduser().resolve()

    def scan_for_assembly(self, directory=None):
        """
        scans the specified directory (default is output/assembly/<sample name>/) for fna file that match the name of the
        sample, then assigns the POSIX path to the self.assembly object attributes
        """

        if directory:
            directory = Path(directory).expanduser().resolve()
        else:
            if file_check(self.output_directory / 'assembly' / self.name):
                directory = self.output_directory / 'assembly' / self.name

        for file in os.scandir(directory):
            if file.name.startswith(self.name + '_') and (fnmatch.fnmatch(file, '*.fasta*') or fnmatch.fnmatch(file, '*.fna*')):
                self.assembly_file = Path(file).expanduser().resolve()
                return

        if self.assembly:
            for file in os.scandir(directory):
                if fnmatch.fnmatch(file, 'assembly.fasta'):
                    self.assembly_file = Path(file).expanduser().resolve()
                    return


def kraken2(sample):
    """
    KRAKEN2 ANALYSIS
    """
    if sample.kraken2:

        mlog.log_section_header('Kraken2 - Species/contamination prediction from read files',
                                colors=sample.colors,
                                stdout_verbosity_level=sample.stdout_verbosity,
                                log_file_verbosity_level=sample.log_file_verbosity,
                                log_file=sample.log_file)
        mlog.log_explanation('Kraken 2 is the newest version of Kraken, a taxonomic classification system using'
                             ' exact k-mer matches to achieve high accuracy and fast classification speeds. '
                             'This classifier matches each k-mer within a query sequence to the lowest common '
                             'ancestor (LCA) of all genomes containing the given k-mer. The k-mer assignments '
                             'inform the classification algorithm.  The output from this analysis currently '
                             'requires manual curation to validate a) species classification and b) '
                             'contamination detection.',
                             colors=sample.colors,
                             stdout_verbosity_level=sample.stdout_verbosity,
                             log_file_verbosity_level=sample.log_file_verbosity,
                             log_file=sample.log_file)

        # scan for Illumina read files
        mlog.log('scanning for input files...',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        sample.scan_for_fastq()
        mlog.log('Current value for R1: ' + mlog.green(str(sample.r1_fastq)),
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        mlog.log('Current value for R2: ' + mlog.green(str(sample.r2_fastq)) + '\n',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)

        if kraken2_precheck(kraken2_database=sample.kraken2_database,
                            force=sample.force,
                            kraken2_directory=sample.kraken2_directory,
                            kraken2_report=sample.kraken2_report,
                            r1_fastq=sample.r1_fastq,
                            r2_fastq=sample.r2_fastq,
                            colors=sample.colors,
                            stdout_verbosity_level=sample.stdout_verbosity,
                            log_file_verbosity_level=sample.log_file_verbosity,
                            log_file=sample.log_file):

            # check that the read files are not compressed
            compression_handler(sample, 'decompress', ['r1_fastq', 'r2_fastq'])
            # run kraken2
            run_kraken2(kraken2_database=sample.kraken2_database,
                        kraken2_report=sample.kraken2_report,
                        r1_fastq=sample.r1_fastq,
                        r2_fastq=sample.r2_fastq,
                        ramdisk=sample.ramdisk,
                        threads=sample.cores,
                        colors=sample.colors,
                        stdout_verbosity_level=sample.stdout_verbosity,
                        log_file_verbosity_level=sample.log_file_verbosity,
                        log_file=sample.log_file)


def assembly(sample):
    """
    ASSEMBLY
    """
    if sample.assembly:

        mlog.log_section_header('Assembly Prechecks',
                                colors=sample.colors,
                                stdout_verbosity_level=sample.stdout_verbosity,
                                log_file_verbosity_level=sample.log_file_verbosity,
                                log_file=sample.log_file)
        mlog.log_explanation('Determine if the requisite conditions have been met to perform adapter trimming with '
                             'bbduk and then assembly using Unicycler. If long read data was provided, preprocess it '
                             'with filtlong.',
                             colors=sample.colors,
                             stdout_verbosity_level=sample.stdout_verbosity,
                             log_file_verbosity_level=sample.log_file_verbosity,
                             log_file=sample.log_file)

        # scan for Illumina read files
        mlog.log('scanning for input files...',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        sample.scan_for_fastq()
        mlog.log('Current value for R1: ' + mlog.green(str(sample.r1_fastq)),
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        mlog.log('Current value for R2: ' + mlog.green(str(sample.r2_fastq)) + '\n',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        # update the assembly file attribute
        sample.scan_for_assembly()
        mlog.log('Current value for assembly is: ' + mlog.green(str(sample.assembly_file)) + '\n',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)

        # perform the assembly precheck to a) validate that an assembly can be performed and b) if that assembly will
        # include long read data
        precheck = assembly_precheck(r1_fastq=sample.r1_fastq,
                                     r2_fastq=sample.r2_fastq,
                                     raw_long_reads_file=sample.raw_long_reads_file,
                                     individual_assembly_directory=sample.individual_assembly_directory,
                                     assembly=sample.assembly_file,
                                     force=sample.force,
                                     colors=sample.colors,
                                     stdout_verbosity_level=sample.stdout_verbosity,
                                     log_file_verbosity_level=sample.log_file_verbosity,
                                     log_file=sample.log_file)

        if not precheck:
            return

        # check that the short read files are not compressed
        compression_handler(sample, 'decompress', ['r1_fastq', 'r2_fastq'])

        # perform adapter trimming
        trim_check = adapter_trimming(r1_fastq=sample.r1_fastq,
                                      r2_fastq=sample.r2_fastq,
                                      r1_fastq_trimmed=sample.r1_fastq_trimmed,
                                      r2_fastq_trimmed=sample.r2_fastq_trimmed,
                                      adapter_file=sample.adapter_file,
                                      force=sample.force,
                                      colors=sample.colors,
                                      stdout_verbosity_level=sample.stdout_verbosity,
                                      log_file_verbosity_level=sample.log_file_verbosity,
                                      log_file=sample.log_file)

        if not trim_check:
            return

        # if precheck = 'long', we will attempt to preprocess the long read data now
        if precheck == 'long':

            sample.processed_long_reads = sample.processed_reads_directory / (sample.name + '_processed_long_reads.fastq')

            long_read_filter = filter_long_reads(r1_fastq=sample.r1_fastq_trimmed,
                                                 r2_fastq=sample.r2_fastq_trimmed,
                                                 raw_long_reads=sample.raw_long_reads,
                                                 processed_long_reads=sample.processed_long_reads,
                                                 force=sample.force,
                                                 colors=sample.colors,
                                                 stdout_verbosity_level=sample.stdout_verbosity,
                                                 log_file_verbosity_level=sample.log_file_verbosity,
                                                 log_file=sample.log_file)

            # determine the long read settings based on the results of the long read filtering
            if long_read_filter:
                long_read_file_for_assembly = sample.processed_long_reads
            else:
                long_read_file_for_assembly = sample.raw_long_reads
        else:
            long_read_file_for_assembly = None

        # perform unicycler assembly
        run_unicycler_assembly(r1_fastq_trimmed=sample.r1_fastq_trimmed,
                               r2_fastq_trimmed=sample.r2_fastq_trimmed,
                               long_read_file=long_read_file_for_assembly,
                               mode=precheck,
                               individual_assembly_directory=sample.individual_assembly_directory,
                               threads=sample.cores)

        # compress the trimmed read files
        compression_handler(sample, 'compress', ['r1_fastq_trimmed', 'r2_fastq_trimmed'])

        # update the assembly file attribute
        sample.scan_for_assembly()

        # copy the unicycler log file to the might log file
        unicycler_log_file = sample.individual_assembly_directory / 'unicycler.log'
        mlog.log('',
                 colors=sample.colors,
                 stdout_verbosity_level=0,
                 log_file_verbosity_level=sample.log_file_verbosity_level,
                 log_file=sample.log_file)
        with open(str(unicycler_log_file), 'r') as uni_log:
            for line in uni_log.readlines():
                mlog.log(str(line),
                         colors=sample.colors,
                         stdout_verbosity_level=0,
                         log_file_verbosity_level=sample.log_file_verbosity_level,
                         log_file=sample.log_file)

        # reformat the assembly file
        compression_handler(sample, 'decompress', ['assembly_file'])

        assembly_file_formatting(assembly_file=(sample.individual_assembly_directory / 'assembly.fasta'),
                                 sample_name=sample.name,
                                 colors=sample.colors,
                                 stdout_verbosity_level=sample.stdout_verbosity,
                                 log_file_verbosity_level=sample.log_file_verbosity,
                                 log_file=sample.log_file)


def amr(sample):
    """
    AMR Prediction
    """
    if sample.amr:

        mlog.log_section_header('Andale AMR Identification',
                                colors=sample.colors,
                                stdout_verbosity_level=sample.stdout_verbosity,
                                log_file_verbosity_level=sample.log_file_verbosity,
                                log_file=sample.log_file)
        mlog.log_explanation('Use Andale to perform AMR Identification based on the available files. ',
                             colors=sample.colors,
                             stdout_verbosity_level=sample.stdout_verbosity,
                             log_file_verbosity_level=sample.log_file_verbosity,
                             log_file=sample.log_file)

        #ascii_art(logo='andale')

        # scan for fastq and assembly files
        mlog.log('Updating read file attributes',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        sample.scan_for_fastq()
        mlog.log('Current value for R1: ' + mlog.green(str(sample.r1_fastq)),
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        mlog.log('Current value for R2: ' + mlog.green(str(sample.r2_fastq)) + '\n',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        mlog.log('Updating assembly file attributes',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)
        sample.scan_for_assembly()
        mlog.log('Current value for assembly: ' + mlog.green(str(sample.assembly_file)) + '\n',
                 colors=sample.colors,
                 stdout_verbosity_level=sample.stdout_verbosity,
                 log_file_verbosity_level=sample.log_file_verbosity,
                 log_file=sample.log_file)

        # determine if the requested analysis type is valid, downgrade or switch if necessary
        sample.amr_analysis_type = amr_precheck(r1_fastq=sample.r1_fastq,
                                                r2_fastq=sample.r2_fastq,
                                                assembly=sample.assembly_file,
                                                individual_amr_directory=sample.individual_amr_directory,
                                                analysis_type=sample.amr,
                                                force=sample.force,
                                                colors=sample.colors,
                                                stdout_verbosity_level=sample.stdout_verbosity,
                                                log_file_verbosity_level=sample.log_file_verbosity,
                                                log_file=sample.log_file)

        # If the amr_precheck() method returned a non-None analysis type, verify the required databases are present and
        # run the analysis
        if sample.amr_analysis_type:

            # verify the databases are installed
            sample.amrfinder_database, sample.ariba_database = database_check(update=sample.update,
                                                                              colors=sample.colors,
                                                                              stdout_verbosity_level=sample.stdout_verbosity,
                                                                              log_file_verbosity_level=sample.log_file_verbosity,
                                                                              log_file=sample.log_file)

            # create the individual_amr_directory
            if not sample.individual_amr_directory.exists():
                subprocess.run(['mkdir', str(sample.individual_amr_directory)])

            # run the validated analyses
            if sample.amr_analysis_type == 'combination' or sample.amr_analysis_type == 'reads':
                if ariba_run(sample,
                             colors=sample.colors,
                             stdout_verbosity_level=sample.stdout_verbosity,
                             log_file_verbosity_level=sample.log_file_verbosity,
                             log_file=sample.log_file):
                    amrfinderplus_run(sample,
                                      ['ariba_assemblies', 'amrfinder_output_from_ariba'],
                                      'reads',
                                      colors=sample.colors,
                                      stdout_verbosity_level=sample.stdout_verbosity,
                                      log_file_verbosity_level=sample.log_file_verbosity,
                                      log_file=sample.log_file)
            if sample.amr_analysis_type == 'combination' or sample.amr_analysis_type == 'contigs':
                amrfinderplus_run(sample,
                                  ['assembly_file', 'amrfinder_output_from_contigs'],
                                  'contigs',
                                  colors=sample.colors,
                                  stdout_verbosity_level=sample.stdout_verbosity,
                                  log_file_verbosity_level=sample.log_file_verbosity,
                                  log_file=sample.log_file)

            # summarize the results
            summarize_amr_finder_output(sample,
                                        colors=sample.colors,
                                        stdout_verbosity_level=sample.stdout_verbosity,
                                        log_file_verbosity_level=sample.log_file_verbosity,
                                        log_file=sample.log_file)


def cleanup(sample):
    """
    CLEANUP
    """
    mlog.log_section_header('Clean up',
                            colors=sample.colors,
                            stdout_verbosity_level=sample.stdout_verbosity,
                            log_file_verbosity_level=sample.log_file_verbosity,
                            log_file=sample.log_file)
    mlog.log_explanation('Analysis is complete. We will now compress large files',
                         colors=sample.colors,
                         stdout_verbosity_level=sample.stdout_verbosity,
                         log_file_verbosity_level=sample.log_file_verbosity,
                         log_file=sample.log_file)

    compression_handler(sample, 'compress', ['r1_fastq', 'r2_fastq'])

    mlog.log_section_header('All done',
                            colors=sample.colors,
                            stdout_verbosity_level=sample.stdout_verbosity,
                            log_file_verbosity_level=sample.log_file_verbosity,
                            log_file=sample.log_file)
    mlog.log_explanation('Thanks for using Might!',
                         colors=sample.colors,
                         stdout_verbosity_level=sample.stdout_verbosity,
                         log_file_verbosity_level=sample.log_file_verbosity,
                         log_file=sample.log_file)

    ascii_art('might')

if __name__ == '__main__':
    premain()


#!/usr/bin/env python
# encoding: utf-8
import os
import subprocess
import time
from pathlib import Path
import pandas as pd
from Bio import SeqIO
import might.might_log as mlog
from might.miscellaneous import compression_handler, suppress_stdout_stderr, file_check


def amr_precheck(r1_fastq=None, r2_fastq=None, assembly=None, individual_amr_directory=None, analysis_type=None,
                 force=False, colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None):
    """
     sets all of the file/directory path attributes associated with an andale run. Analysis type is assigned by the
     user in the CLI, then validated by performing scan_for_fastq() and scan_for_assembly(),
    """
    """
    Determination of the type of andale analysis to run. There are 4 primary options: 
    --combination: reads using ARIBA and AMRFinderPlus on the ARIBA output and the contigs file
    --reads: reads only using ARIBA and AMRFinderPlus
    --contigs: AMRFinderPlus on the contigs file only
    --summary_only: which performs no new analysis but instead summarizes the preexisting data 
    """

    # if the user specifies that they want --summary_only, this will supersede any other requests
    if analysis_type == "summary":
        amr_analysis_type = "summary"
        return amr_analysis_type

    # for all other cases, we will use the user input (if it exists) and/or the presence/absence of the input

    # check to see if the AMR output is already present.
    if individual_amr_directory:
        if file_check(individual_amr_directory):
            if force:
                mlog.log('Andale output already exists for this sample, but --force was invoked so existing output will'
                         ' be overwritten\n',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                subprocess.run(['rm', '-r', str(individual_amr_directory)])
            # else:
            #    log.log('Andale output for this sample was already present, so this analysis is being aborted\n')
            #    return None

    # case #1: all three input files are present
    if r1_fastq and r2_fastq:
        if assembly:
            if analysis_type == 'ARIBA only':
                mlog.log('Andale analysis type has been set to: reads only',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                amr_analysis_type = 'reads'
            if analysis_type == 'contigs only':
                mlog.log('Andale analysis type has been set to: contigs only',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                amr_analysis_type = 'contigs'
            else:
                mlog.log('Andale analysis type has been set to: combination',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                amr_analysis_type = 'combination'

    # case #2: only the read files are present
    if r1_fastq and r2_fastq and not assembly:
        mlog.log('Andale analysis type has been set to: reads only',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        amr_analysis_type = 'reads'

    # case #3: only the assembly file is present
    if (not r1_fastq or not r2_fastq) and assembly:
        mlog.log('Andale analysis type has been set to: contigs only',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        amr_analysis_type = 'contigs'

    # case #4: none of the requisite file types are present so no analysis can be performed
    if (not r1_fastq or not r2_fastq) and not assembly:
        mlog.log('Insuffcient files available for any AMR prediction, aborting this analysis',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return None

    return amr_analysis_type


def ariba_run(sample, colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None):
    """
    This method will handle the ariba/ncbi run command for a single sample:
    1. Check/modify compression of read files
    2. Perform local assembly of the read files using ariba and the AMRFinderPlus-derived prepared reference database
    3. Compress the read files once ariba is done
    """

    mlog.log_section_header('ARIBA - AMR prediction from read files',
                            colors=colors,
                            stdout_verbosity_level=stdout_verbosity_level,
                            log_file_verbosity_level=log_file_verbosity_level,
                            log_file=log_file)
    mlog.log_explanation('ARIBA is used to assemble raw illumina reads to the AMRFinderPlus reference database prepared '
                        'during the AMR Database Validation step. While ARIBA offers its own built in prediction method'
                        ' (refer to /AMR/<sample name>/ARIBA_output/report.tsv), we will be using the assemblies.fa '
                        'file as input for AMRFinderPlus protein-based prediction',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)

    # check for checkpoint file
    checkpoint_file = sample.ariba_assemblies

    if file_check(checkpoint_file):
        mlog.log('CHECKPOINT DETECTED: ariba checkpoint file detected, skipping this analysis',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return True

    # decompress the R1 and R2 reads (if compressed)
    compression_handler(sample, 'decompress', ['r1_fastq', 'r2_fastq'])
    time.sleep(1)

    """
    ariba is prone to erroring out when running mutliple threads. Failure of the process is indictated by the message: "Stopping!
    Signal received: 28". We suspect this is a terminal error relating to timeout. In an effort to get around this, we will
    loop ariba until it completes without returning the error code with a maximum of 3 attempts
    """

    # run ariba, directing output files to sample.ariba_output_directory and piping the stderr to ariba_process
    ariba_cmd = 'ariba run ' + str(sample.ariba_database) + '/ ' + str(sample.r1_fastq) + ' ' + \
                str(sample.r2_fastq) + ' ' + str(sample.ariba_output_directory)

    mlog.log('Command: ' + mlog.bold(ariba_cmd) + '\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    error_code_received = True
    attempt_count = 0
    formatted_attempt_count = ['0', '1st', '2nd', '3rd']
    while error_code_received and attempt_count < 4:
        attempt_count += 1

        mlog.log('Running ARIBA...',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

        ariba_process = subprocess.run(['ariba', 'run',
                                        str(sample.ariba_database),
                                        str(sample.r1_fastq),
                                        str(sample.r2_fastq),
                                        str(sample.ariba_output_directory),
                                        '--threads', sample.cores],
                                       stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)

        # read in the stderr captured in ariba_process
        ariba_error = str(ariba_process.stderr)

        # check to see if the aborted run signal is present in ARIBA's stderr. If so, delete the failed ariba output
        # and try again
        if "Signal received: 28" in ariba_error:
            mlog.log('ariba runtime error encountered, resulting in run termination. Retrying for the ' +
                    formatted_attempt_count[attempt_count] + ' time',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)

            subprocess.run(['rm', '-r', sample.ariba_output_directory])
        else:
            error_code_received = False

            mlog.log('ARIBA run finished successfully',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)

            return True

    return False


def amrfinderplus_run(sample, list_of_attributes, analysis_type, colors=1, stdout_verbosity_level=1,
                      log_file_verbosity_level=1, log_file=None):
    """
    run AMRFinderPlus on the selected sample object for the attribute (file) selected (contig file/ariba assemblies file)
    list_of_attributes is a two item list [input file for AMRFinderPlus, output file from AMRFinderPlus]
    """

    mlog.log_section_header('AMRFinderPlus - AMR prediction from assemblies for ' + analysis_type,
                            colors=colors,
                            stdout_verbosity_level=stdout_verbosity_level,
                            log_file_verbosity_level=log_file_verbosity_level,
                            log_file=log_file)
    mlog.log_explanation('AMRFinderPlus will be used to perform AMR prediction using the assemblies associated with'
                        ' ' + analysis_type + '.',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)

    output_file = str(getattr(sample, list_of_attributes[1]))

    # check for output file (checkpoint)
    if file_check(output_file):
        mlog.log('CHECKPOINT DETECTED: checkpoint file detected, skipping this analysis',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return

    # decompress the input file, if required
    mlog.log('decompressing the input file (if necessary)\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)
    compression_handler(sample, 'decompress', [list_of_attributes[0]])
    time.sleep(2)

    input_file = str(getattr(sample, list_of_attributes[0]))

    # run amrfinder in nucleotide mode for assembled_genes.fa
    amrfinderplus_cmd = 'amrfinder --plus -d ' + str(sample.amrfinder_database) + ' -o ' + str(output_file) + ' -n ' +\
                        str(input_file) + ' --threads 1'

    mlog.log('Command: ' + mlog.bold(amrfinderplus_cmd) + '\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)
    mlog.log('Now running AMRFinder for ' + analysis_type,
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    subprocess.run(['amrfinder',
                    '--plus',
                    '-d', str(sample.amrfinder_database),
                    '-o', str(output_file),
                    '-n',
                    str(input_file),
                    '--threads', '1'],
                   stderr=subprocess.STDOUT,
                   stdout=subprocess.DEVNULL)

    # compress input file to save space
    mlog.log('AMRFinder finished for ' + analysis_type + '.',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)
    compression_handler(sample, 'compress', [list_of_attributes[0]])

    return


def summarize_amr_finder_output(sample, colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None):
    """
    summarize all of the amr_finder generated output for all of the samples into a single file, andale_summary.csv
    Merge all of the available AMRFinderPlus output derived from contigs, repeat the process for output based on ariba
    assemblies. Generate TWO integrated report files:
        1) andale_output_raw.csv, which is merely a concatenation of all available output for every sample, with true
           duplicates removed
        2) andale_output_FILTERED.csv, TBD
    """

    mlog.log_section_header('Summarize Results',
                            colors=colors,
                            stdout_verbosity_level=stdout_verbosity_level,
                            log_file_verbosity_level=log_file_verbosity_level,
                            log_file=log_file)
    mlog.log_explanation('Generate summary tables of all AMRFinderPlus results.',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)

    ariba_summary_file = sample.individual_amr_directory / 'ariba_summary.csv'
    contig_summary_file = sample.individual_amr_directory / 'contigs_summary.csv'
    andale_output_raw_file = sample.individual_amr_directory / 'andale_output_raw.csv'
    andale_output_filtered = sample.individual_amr_directory / 'andale_output_filtered.csv'

    # remove existing summary files is they are present
    for file in [ariba_summary_file, contig_summary_file, andale_output_raw_file, andale_output_filtered]:
        if Path(file).resolve().exists():
            subprocess.run(['rm', str(Path(file))])

    # Create DataFrame "andale_output_raw" which we will add all of the data generated during the analysis run to
    andale_output_raw = pd.DataFrame(
        columns=['sample', 'method', 'Gene symbol', 'Sequence name', 'Method', '% Coverage of reference sequence',
                 '% Identity to reference sequence', 'Start', 'Stop', 'Strand', 'Contig id',
                 'Accession of closest sequence', 'Name of closest sequence', 'Class', 'Subclass'])

    for source in ['ariba', 'Contigs']:

        if source == 'ariba':
            input_file = str(sample.amrfinder_output_from_ariba)
            summary_file = ariba_summary_file

        if source == 'Contigs':
            input_file = str(sample.amrfinder_output_from_contigs)
            summary_file = contig_summary_file
            if Path(input_file).exists():
                # v0.4 -- if contigs were used, add the total length of the contig the hit is located on to the output
                # csv(s)
                contig_lengths = get_contig_lengths(sample)

        if Path(input_file).exists():

            mlog.log('Generating summary file: ' + mlog.bold(input_file),
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)

            input_column_list = ['Gene symbol', 'Sequence name', 'Scope', 'Element type', 'Element subtype',
                                 'Class', 'Subclass', 'Method', '% Coverage of reference sequence',
                                 '% Identity to reference sequence', 'Start', 'Stop', 'Strand', 'Contig id',
                                 'Accession of closest sequence', 'Name of closest sequence']
            output_column_list = ['sample', 'source', 'Gene symbol', 'Sequence name', 'Method',
                                  '% Coverage of reference sequence', '% Identity to reference sequence', 'Start',
                                  'Stop', 'Strand', 'Contig id', 'Contig length', 'Accession of closest sequence',
                                  'Name of closest sequence', 'Scope', 'Element type', 'Element subtype', 'Class',
                                  'Subclass']

            # Read in the current input file (either contig or ariba output) to the DataFrame "report_data"
            report_data = pd.read_csv(input_file, sep='\t', usecols=input_column_list)
            report_data['sample'] = sample.name
            report_data['source'] = source

            # v0.4 -- if contigs were used, add the total length of the contig the hit is located on to "report_data"
            if source == 'Contigs':
                report_data['Contig id'] = report_data['Contig id'].astype(str)
                report_data = report_data.merge(contig_lengths, on='Contig id', how='left')

            # add the current "report_data" to "andale_output_raw"
            andale_output_raw = pd.concat([andale_output_raw, report_data], ignore_index=True, sort=False)

            # v0.5 -- pandas to_csv passes a FutureWarning we don't need to see, so we'll redirect the terminal output
            # to devnull temporarily
            with suppress_stdout_stderr():

                # add all of the sample data for a given source (contigs or ariba) to the corresponding summary file
                if not os.path.isfile(summary_file):
                    report_data.to_csv(summary_file, sep=',', index=False, columns=output_column_list, mode='a')
                else:
                    report_data.to_csv(summary_file, sep=',', index=False, columns=output_column_list, mode='a',
                                       header=False)

    # Generate andale_output_raw.csv by concatenating ariba_summary.csv and contigs_summary.csv

    mlog.log('Generating summary file: ' + mlog.bold(str(andale_output_raw_file)),
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    andale_output_raw.to_csv(andale_output_raw_file, sep=',', index=False, columns=output_column_list, mode='a')

    # Results filtering: Since we are potentially calling genes from two sources (assembled contigs and ARIBA output) we
    # anticipate that there will be significant redundancy in the output. In an attempt not to drive folks crazy with
    # deduplicating the resulting .csv file, we will perform some "safe" deduplication operations on the output
    # dataframe. Note that you still have access to the complete dataset in the andale_output_raw.csv file

    # First, we can safely drop all true duplicate entries from the dataframe
    andale_output_raw.drop_duplicates(inplace=True)

    # Second, in the case where the result for a given sample/gene is identical in the method of call (either EXACTX or
    # ALLELEX) as well as the %coverage and %ID, we will selectively keep the result from the contigs as this will
    # provide the greatest context
    filtered_andale_output = andale_output_raw.groupby(['sample', 'Gene symbol', 'Method',
                                                        '% Coverage of reference sequence',
                                                        '% Identity to reference sequence'])
    for name, group in filtered_andale_output:
        if len(group.index) > 1:
            if 'Contigs' in group.source.values:
                filter_methods = ['EXACTX', 'ALLELEX']
                filtered_group = group[((group['Method'].isin(filter_methods)) |
                                       ((group['Method'] == 'BLASTX') &
                                        (group['% Coverage of reference sequence'] >= 90) &
                                        (group['% Identity to reference sequence'] >= 90))) &
                                        (group['source'] == 'Contigs')]
                for i in group.index:
                    if i not in filtered_group.index:
                        andale_output_raw.drop(i, inplace=True)

    # Third, in the case where an EXACTX or ALLELEX hit OR a BLASTX hit with >95/95 %coverage/%ID exists for a given
    # sample/gene combination, we will remove any PARTIAL_CONTIG_ENDX or PARTIAL_CONTIGX that are the result of ARIBA
    # assembling partial genes adjacent to the target gene
    filtered_andale_output = andale_output_raw.groupby(['sample', 'Gene symbol'])
    for name, group in filtered_andale_output:
        if len(group.index) > 1:
            if ('PARTIAL_CONTIG_ENDX' in group.Method.values) or ('PARTIAL_CONTIGX' in group.Method.values):
                filter_methods = ['EXACTX', 'ALLELEX']
                filtered_group = group[(group['Method'].isin(filter_methods)) |
                                       ((group['Method'] == 'BLASTX') &
                                        (group['% Coverage of reference sequence'] >= 90) &
                                        (group['% Identity to reference sequence'] >= 90))]
                print(filtered_group.index)
                for i in group.index:
                    if i not in filtered_group.index:
                        andale_output_raw.drop(i, inplace=True)

    andale_output_raw.to_csv(andale_output_filtered, sep=',', index=False, columns=output_column_list, mode='a')

    return

def get_contig_lengths(sample):
    """
    method takes a multifasta file and returns a pandas series where the index is the name of the contigs
    and the values are the lengths of those contigs
    """

    compression_handler(sample, 'decompress', ['assembly_file'])

    # Use SeqIO to read all of the contig names and lists into two lists, contig_names and contig_lengths
    contig_length_dict = {'Contig id': [], 'Contig length': []}
    for contig in SeqIO.parse(sample.assembly_file, "fasta"):
        contig_length_dict['Contig id'].append((str(contig.id).split(' '))[0])
        contig_length_dict['Contig length'].append(int(len(contig.seq)))

    # convert the two lists into a series object
    contig_lengths = pd.DataFrame(contig_length_dict, columns=['Contig id', 'Contig length'])

    compression_handler(sample, 'compress', ['assembly_file'])

    return contig_lengths




#!/usr/bin/env python
# encoding: utf-8
import copy
import multiprocessing as mp
import os
import subprocess
import sys
from fnmatch import fnmatch
from math import ceil
from multiprocessing import Pool
from pathlib import Path
import might.might_log as mlog
from might.database_methods import database_check
from might.miscellaneous import parse_arguments, check_for_dependency, quit_with_error, validate_sample_sheet, \
    ramdisk_handler, round_down, ascii_art


def main():

    ascii_art('might')

    args = parse_arguments(allmight=True)

    # make sure that the output directory exists
    args.output_directory = Path(args.output).expanduser().resolve()
    if not args.output_directory.exists():
        subprocess.run(['mkdir', str(args.output_directory)])

    # make sure that the reads/raw_reads subdirectory of the output directory exists
    raw_reads_directory = args.output_directory / 'reads' / 'raw_reads'
    if not raw_reads_directory.exists():
        subprocess.run(['mkdir', '-p', str(raw_reads_directory)])

    # a sample list should have been generated during parse_arguments. Set it to sample_list for use here
    sample_list = args.sample_list

    # If bcl2fastq2 analysis is requested, parse_arguments() should have validated the sample sheet, copied it to the
    # required directory and produced a list of the samples we are expecting to have data for.
    if args.bcl2fastq:
        sample_list = bcl2fastq2(sample_sheet=args.sample_sheet,
                                 run_directory=args.run_directory,
                                 sample_list=args.sample_list,
                                 cores=args.cores)

        # Sanity check: make sure that we got samples back bcl2fastq2
        if len(sample_list) == 0:
            quit_with_error('ERROR: bcl2fastq2 returned no samples for analysis!')

        # We are going to move the fastq files for the good samples to the reads/raw_reads subdirectory of the output
        # directory. If the reads for a given sample are already present, we will abort analysis of this sample and
        # report that to the user.

        for sample in sample_list:
            for file in os.scandir(str(raw_reads_directory)):
                if file.name.startswith(sample + '_'):
                    mlog.log('WARNING: ' + sample + ' already has read files present in the output/reads/raw_reads '
                            'directory! This sample will be omitted from analysis.')
                    sample_list.remove(sample)

        # Sanity check: make sure that we still have samples to analyze
        if len(sample_list) == 0:
            quit_with_error('There are no new samples for analysis! Please remove old read files from the output '
                            'directory, or specify a new output directory')

        for sample in sample_list:
            for file in os.scandir(str(Path(args.run_directory).expanduser().resolve() / 'Data' / 'Intensities' / 'BaseCalls')):
                if file.name.startswith(sample + '_'):
                    subprocess.run(['mv', file, str(raw_reads_directory)])

        # If no other analyses were requested, we will exit here. All that follows pertains to individual parallel
        # analyses
        if not args.kraken2 and not args.assembly and not args.amr:
            mlog.log('Since only bcl2fastq was requested, AllMight.py is finished. Your fastq files can be found here'
                    ':\n' + str(raw_reads_directory))
            sys.exit()

    # If kraken2 analysis was requested by the user, lets go ahead and mount the ramdisk now
    if args.kraken2:
        if not ramdisk_handler(mount=True,
                               ramdisk_directory=args.ramdisk,
                               directory_to_load=args.kraken2_database):
            args.ramdisk=None

    # If amr analysis was requested or a database update was requested, lets verify or install the amr databases now
    if args.amr:
        database_check(update=args.update)

    # Initialize parallel processing of the samples
    mlog.log_section_header('Starting Individual Analysis')
    mlog.log_explanation('Sample objects will be processed in parallel based on the parameters given')

    mp.set_start_method('spawn')
    pool_size = int(round_down(int(args.cores), 4) / 4)

    if int(args.cores) > 4:
        args.cores = '4'

    sample_arg_list = queue_loader(sample_list=sample_list, args=args, raw_reads_directory=raw_reads_directory)

    individual_samples = []

    for entry in sample_arg_list:
        individual_samples.append(Might.IndividualSequencingSample(entry))

    # run kraken2 analysis for all samples in parallel if requested
    if args.kraken2:
        p = Pool(pool_size)
        p.map(Might.kraken2, individual_samples)
        p.close()
        p.join()

    # run assembly analysis for all samples in parallel
    if args.assembly:
        p = Pool(pool_size)
        p.map(Might.assembly, individual_samples)
        p.close()
        p.join()

    # run amr analysis for all samples in parallel
    if args.amr:
        p = Pool(pool_size)
        p.map(Might.amr, individual_samples)
        p.close()
        p.join()

    # Clean up
    p = Pool(pool_size)
    p.map(Might.cleanup, individual_samples)
    p.close()
    p.join()

    ramdisk_handler(unmount=True, ramdisk_directory=str(args.ramdisk))


def bcl2fastq2(sample_sheet=None, run_directory=None, sample_list=None, cores='1'):
    """
    run bcl2fastq2 using SampleSheet.csv and the path to the run directory it resides in
    """

    mlog.log_section_header('bcl2fastq2')
    mlog.log_explanation('bcl2fastq2 demultiplexes sequencing data and converts base call (BCL) files into fastq files')

    if not sample_sheet or not run_directory:
        quit_with_error('bcl2fastq2 requires both a valid sample sheet and the path to the Illumina run '
                        'directory!')

    # determine if bcl2fastq2 is in PATH
    if not check_for_dependency('bcl2fastq'):
        quit_with_error('Unable to locate bcl2fastq2 in your path!')

    # bcl2fastq2 automatically generates fastq files in the [run directory]/Data/Intensities/BaseCalls directory
    basecalls_directory = Path(run_directory).expanduser().resolve() / 'Data' / 'Intensities' / 'BaseCalls'

    # make sure we are not overwriting any existing files. To do this, we will scan the basecalls_directory for any
    # fastq.gz files. If any are present, we will abort and prompt the user to clean this directory of any existing
    # bcl2fastq2 output

    for file in os.scandir(str(basecalls_directory)):
        if fnmatch(file, '*fastq*'):
            quit_with_error('There are already fastq files in the BaseCalls directory! If you wish to perform '
                            'bcl2fastq on this run directory, you will need to remove all fastq files!')

    # make sure that a sample_list was passed to the method, as this signals that the sample_sheet file was already
    # validated during parse_arguments(). If sample_list is None, we will run the validate_sample_sheet() now.

    if not sample_list:

        if not Path(run_directory).expanduser().resolve().is_dir():
            quit_with_error('User requested bcl2fastq2 be performed, but the path to the Illumina run output '
                            'directory provided by the user cannot be located.\n Please ensure that the path to the'
                            ' Illumina run output directory provided using --run-directory is correct!')
        if not Path(sample_sheet).expanduser().resolve().exists():
            quit_with_error('User requested bcl2fastq2 be performed, but the path provided to the SampleSheet.csv '
                            'file can not be verified.\n Please ensure that the path to the SampleSheet.csv file '
                            'provided using --sample-sheet is correct!')
        # validate the formatting of the SampleSheet.csv file before proceeding
        sample_list, sample_sheet = validate_sample_sheet(samplesheet=sample_sheet, run_directory=run_directory)

    # At this point, we should have a valid sample_sheet file in the run_directory as well as a BaseCalls subdirectory
    # that is free of prior bcl2fastq2 results. Time to boogie.

    # Determine the appropriate thread allocations for the run
    # --loading-threads: default is 4. Only change if the number of allotted processors is <4
    if int(cores) < 4:
        loading_threads = str(cores)
    else:
        loading_threads = '4'
    # --processing_threads: default is... as many as you are willing to give it.
    processing_threads = str(cores)
    # --writing threads: default is 4. Only change if the number of allotted processors is <4
    if int(cores) < 4:
        writing_threads = str(cores)
    else:
        writing_threads = '4'

    # prepare the bcl2fastq2 command as a string for logging purposes
    bcl2fastq2_cmd = 'bcl2fastq --runfolder-dir ' + str(Path(run_directory).expanduser().resolve()) + \
                     ' --loading-threads ' + loading_threads + ' --processing-threads ' + processing_threads + \
                     ' --writing-threads ' + writing_threads + ' --barcode-mismatches=1 --no-lane-splitting'

    mlog.log('Command: ' + mlog.bold(bcl2fastq2_cmd) + '\n')

    mlog.log('Running bcl2fastq...')

    subprocess.run(['bcl2fastq',
                    '--runfolder-dir', str(Path(run_directory).expanduser().resolve()),
                    '--loading-threads', loading_threads,
                    '--processing-threads', processing_threads,
                    '--writing-threads', writing_threads,
                    '--barcode-mismatches=1',
                    '--no-lane-splitting'],
                   stderr=subprocess.DEVNULL)

    mlog.log('bcl2fastq2 finished\n')

    # verify that bcl2fastq2 produced at least some fastq files
    bcl2fastq2_status = False
    for file in os.scandir(str(basecalls_directory)):
        if fnmatch(file, '*fastq*'):
            bcl2fastq2_status = True
            break

    if not bcl2fastq2_status:
        quit_with_error('bcl2fastq2 failed to produce any fastq.gz files in the expected location!')

    # verify that bcl2fastq2 produced at least some non 'Undetermined' files, otherwise exit with an error message
    # indicating there is likely a problem with the supplied sample sheet

    bcl2fastq2_status = False
    for file in os.scandir(str(basecalls_directory)):
        if not file.name.startswith('Undetermined') and fnmatch(file, '*fastq*'):
            bcl2fastq2_status = True
            break

    if not bcl2fastq2_status:
        quit_with_error('bcl2fastq2 only produced Undetermined files! Make sure that the sample sheet that you provided'
                        ' is the correct one for the run being analyzed')

    # Determine which samples we successfully generated fastq files for. Determine the size of the fastq output files
    # for each sample. If the files are below a certain size threshold, we will report them as sequencing failures to
    # the user and omit them from any downstream analysis.

    passed_sample_list = []

    for sample in sample_list:
        passed_reads = []
        for file in os.scandir(str(basecalls_directory)):
            if file.name.startswith(sample + '_') and fnmatch(file, '*R1*fastq*') and os.path.getsize(file)>20000000:
                passed_reads.append(file)
            if file.name.startswith(sample + '_') and fnmatch(file, '*R2*fastq*') and os.path.getsize(file)>20000000:
                passed_reads.append(file)
        if len(passed_reads) == 2:
            passed_sample_list.append(sample)

    for sample in sample_list:
        if sample not in passed_sample_list:
            mlog.log('WARNING: ' + mlog.bold(sample) + ' failed to produce quality read files and will be omitted from'
                    ' further analysis')

    # Acceptable losses filter. If more than 10% of isolates fail to pass the initial read size filter, quit with an
    # error asking the user to review the sample sheet for errors. This filter can be overridden with the --no-filters
    # option at the command line

    failure_threshold = ceil(len(sample_list)*0.1)
    failure_count = len(sample_list) - len(passed_sample_list)

    if failure_count > failure_threshold:
        quit_with_error('More than 10% of samples from this run failed to meet to the criteria for analysis. Please '
                        'review the sample sheet as well as library quantification data to ensure that this level of '
                        'sample failure is expected. If so, you can rerun the analysis with the --no-filter option')

    mlog.log('bcl2fastq complete!')
    # Return the list of samples that passed the size threshold for individual analysis
    return passed_sample_list


def queue_loader(sample_list=None, args=None, raw_reads_directory=None):

    sample_arg_list = []
    for sample in sample_list:
        sample_args = copy.copy(args)
        sample_args.sample_name = sample
        sample_args.fastq = str(raw_reads_directory)
        sample_args.fasta = None
        sample_args.verbosity = args.verbosity
        sample_args.log_file_verbosity = 4
        sample_arg_list.append(sample_args)
    return sample_arg_list


if __name__ == '__main__':
    main()




#!/usr/bin/env python
# encoding: utf-8

import argparse
import os
from contextlib import redirect_stderr, redirect_stdout, contextmanager
from filecmp import dircmp
from pathlib import Path
import sys
import csv
import distutils.spawn
import subprocess
import time
import might.might_log as mlog



def file_check(file_path):
    """
    checks to see if the file associated with the individual_sequencing_sample object exists in the file system at the
    specified PATH. Returns True if yes and False if no. Also return False if the attribute does not exist
    """
    if Path(str(file_path)).expanduser().resolve().exists():
        return True
    else:
        return False


def check_for_dependency(program):
    """
    check_for_dependency() will use distutils.spawn to determine if the input dependency is available on the users PATH
    returns True if the dependency is located, False otherwise
    """
    if distutils.spawn.find_executable(str(program)):
        return True
    else:
        return False


def quit_with_error(message):
    """
    Displays the given message and ends the program's execution.
    """
    mlog.log('Error: ' + message, 0, stderr=True)
    sys.exit(1)


def compression_handler(sample, compress_or_decompress, list_of_attributes):
    """
    method handles the compression (gzip) or decompression (gunzip) of files
    """
    # If 'decompress' is requested, check compression state and decompress if necessary
    if compress_or_decompress == 'decompress':
        for attribute in list_of_attributes:
            file_path = str(getattr(sample, attribute))
            if file_path.endswith(".gz"):
                mlog.log('Extracting: ' + mlog.bold(file_path))
                subprocess.run(['gunzip', file_path])
                setattr(sample, attribute, Path(file_path[:-3]))
    # If 'compress' is requested, check compression state and compress if necessary
    if compress_or_decompress == 'compress':
        for attribute in list_of_attributes:
            file_path = str(getattr(sample, attribute))
            if not file_path.endswith(".gz"):
                mlog.log('Compressing: ' + mlog.bold(file_path))
                subprocess.run(['gzip', file_path])
                setattr(sample, attribute, Path(file_path + '.gz'))

    return


def parse_arguments(allmight=False, might=False):
    """
    capture user instructions from CLI. Based on the input provided, initialize an analysis
    """

    # parse user input
    parser = argparse.ArgumentParser(description='MIGHT! MRSN Integrated Genome Handling Tool \n', add_help=False)

    # Required arguments for a given sample. Need to know where to store the results (--output) and the name of the
    # sample (--sample_name)
    required_args = parser.add_argument_group('Required arguments')
    required_args.add_argument('--output',
                               type=str,
                               required=True,
                               help='path to the directory where output is/will be stored')

    # If parse_arguments is being called by the wrapper script AllMight.py, we will include the optional bcl2fastq
    # handling arguments
    if allmight:
        bcl2fastq_args = parser.add_argument_group('bcl2fastq2 arguments')
        bcl2fastq_args.add_argument('--bcl2fastq',
                                    action='store_true',
                                    help='Run bcl2fastq2 to generate demultiplexed fastq files from the bcl files')
        bcl2fastq_args.add_argument('--run-directory',
                                    type=str,
                                    default=None,
                                    help='Path to the run directory to be analyzed')
        bcl2fastq_args.add_argument('--sample-sheet',
                                    type=str,
                                    default=None,
                                    help='Path to the Illumina sample sheet file for the run being analyzed')

    # If parse_arguments is being called by the the individual sample handler Might.py, we will include the optional
    # path handling arguments used to tell the script where to find read/assembly files
    if might:
        input_args = parser.add_argument_group('Input arguments')
        input_args.add_argument('--sample-name',
                                type=str,
                                default=None,
                                help='Name of the sample to be analyzed.')
        input_args.add_argument('--fastq',
                                type=str,
                                help='path to the directory containing the read files for this sample '
                                     '[output/reads/raw_reads]')
        input_args.add_argument('--fasta',
                                type=str,
                                default=None,
                                help='path to the directory containing the assembly file for this sample '
                                     '[output/assembly]')
        input_args.add_argument('--long-reads',
                                type=str,
                                default=None,
                                help='path to the long reads file for this assembly')

    # Analysis arguments governing which analyses are going to be run on this sample.
    analysis_args = parser.add_argument_group('Analysis arguments')
    analysis_args.add_argument('--all',
                               action='store_true',
                               help='run all analysis options')
    analysis_args.add_argument('--kraken2',
                               action='store_true',
                               help='run Kraken2 on read files to determine species ID and potentially detect '
                                    'contamination')
    analysis_args.add_argument('--assembly',
                               action='store_true',
                               help='trim and filter reads using bbduk, then perform assembly using Unicycler')
    analysis_args.add_argument('--amr',
                               choices=['combination', 'reads', 'contigs', 'summary'],
                               help='run Andale using one of the four setting choices')
    analysis_args.add_argument('--mlst',
                               action='store_true',
                               help='perform MLST assignments for samples using MLST')
    analysis_args.add_argument('--plasmidfinder',
                               action='store_true',
                               help='run Plasmidfinder on contig files to identify rep gene content')

    # Resource arguments that provide the script with paths to databases and files required for some analysis steps
    resource_args = parser.add_argument_group('Resource arguments')
    resource_args.add_argument('--kraken2-database',
                               type=str,
                               default=None,
                               help='Path to the kraken2 database. Required for kraken2 analysis')
    resource_args.add_argument('--adapter-file',
                               type=str,
                               default=None,
                               help='Path to the adapter.fa file required for adapter trimming of Illumina reads')
    resource_args.add_argument('--ramdisk',
                               type=str,
                               help='Path to the ramdisk for speeding up kraken2')

    # Optional arguments that modify how the analysis steps are performed
    optional_args = parser.add_argument_group('Optional arguments')
    optional_args.add_argument('--update',
                               action='store_true',
                               help='update AMRFinderPlus and MLST databases')
    optional_args.add_argument('--force',
                               action='store_true',
                               help='force overwrite of existing data/output related to this sample')
    optional_args.add_argument('--cores',
                               type=int,
                               default=1,
                               help='the MAXIMUM number of CPUs to use in the analysis [1]')
    optional_args.add_argument('--verbosity',
                               type=int,
                               default=1,
                               help='the level of reporting done to the terminal window [1]')

    # Help and version arguments
    help_args = parser.add_argument_group('Help')
    help_args.add_argument('-h',
                           '--help',
                           action='help',
                           default=argparse.SUPPRESS,
                           help='show this help message and exit \n')

    # collect the user arguments
    args = parser.parse_args()

    # validate that no conflicting input was provided

    # bcl2fastq2 by definition does NOT run on single samples. In order for a bcl2fastq2 to run, a sample sheet
    # must have been provided, as well as a run directory. Missing either is fatal
    if allmight:
        if args.bcl2fastq:
            if not args.run_directory:
                quit_with_error('User requested bcl2fastq2 be performed, but failed to specify the path to the Illumina'
                                ' run output directory.\n Please ensure that the full path to the Illumina run output '
                                'directory is provided using --run-directory in the future!')
            if not Path(args.run_directory).expanduser().resolve().is_dir():
                quit_with_error('User requested bcl2fastq2 be performed, but the path to the Illumina run output '
                                'directory provided by the user cannot be located.\n Please ensure that the path to the'
                                ' Illumina run output directory provided using --run-directory is correct!')
            if not args.sample_sheet:
                quit_with_error('User requested bcl2fastq2 be performed, but failed to specify the path to the '
                                'SampleSheet.csv file.\n Please ensure that the full path to the SampleSheet.csv file '
                                'is provided using --sample-sheet in the future!')
            if not Path(args.sample_sheet).expanduser().resolve().exists():
                quit_with_error('User requested bcl2fastq2 be performed, but the path provided to the SampleSheet.csv '
                                'file can not be verified.\n Please ensure that the path to the SampleSheet.csv file '
                                'provided using --sample-sheet is correct!')
        # validate the formatting of the SampleSheet.csv file before proceeding
        args.sample_list, args.sample_sheet = validate_sample_sheet(samplesheet=args.sample_sheet,
                                                                    run_directory=args.run_directory)

    # if --all was selected, set the values of all analysis flags to true
    if args.all == True:
        args.bcl2fastq = True
        args.kraken2 = True
        args.assembly = True
        args.amr = True
        #args.mlst = True
        #args.plasmidfinder = True

    # if kraken2 analysis was requested but no kraken2_database was specified, quit with error message
    if args.kraken2 and not args.kraken2_database:
        quit_with_error('User requested kraken2 analysis but failed to provide a kraken2_database path.\nPlease ensure'
                        'that the full path to the prepared kraken2 database is provided using --kraken2_database in '
                        'the future!')

    # if assembly analysis is requested but no adapter.fa file path is specified, quit with error message
    if args.assembly:
        if not args.adapter_file:
            quit_with_error('User requested an assembly but failed to provide the path to the adapter.fa file '
                            'required to perform adapter trimming of the Illumina read files.\nPlease ensure that the '
                            'full path to the adapter.fa file is provided using --adapter_file in the future!')
        if not Path(args.adapter_file).expanduser().resolve().exists():
            quit_with_error('User requested an assembly but the path to the adapter.fa file required to perform adapter'
                            ' trimming of the Illumina read files can not be resolved. \nPlease ensure that the full '
                            'path to the adapter.fa file provided using --adapter_file is correct!')
        if might:
            if args.fasta:
                quit_with_error('User requested assembly be performed, but also provided the path to an assembly file. '
                                'Please provide EITHER --fasta OR --assembly in the future!')

    # validate the number of cores specified by the user.
    if args.cores < 0:
        quit_with_error('User requested a negative number of CPUs.')
    if args.cores > len(os.sched_getaffinity(0)):
        quit_with_error('User requested more cores for this analysis than are currently available.\nThe number of'
                        ' currently available cores is: ' + str(len(os.sched_getaffinity(0))))

    return args


def validate_sample_sheet(samplesheet=None, run_directory=None):
    """
     This method will make sure that:
         1)  the path to the sample sheet is valid
         2)  the sample sheet has the proper name (SampleSheet.csv) that bcl2fastq2 will accept
         3)  the sample sheet is free of duplicate sample names and/or barcode combinations
     Method will return a list of sample names
     """
    # sanitize input
    samplesheet = Path(samplesheet).expanduser().resolve()
    run_directory = Path(run_directory).expanduser().resolve()

    # determine if the run_directory can be located at the provided path
    if not run_directory.is_dir():
        quit_with_error('User requested bcl2fastq2 be performed, but the path to the Illumina run output directory '
                        'provided by the user cannot be located.\n Please ensure that the path to the Illumina run '
                        'output directory provided using --run-directory is correct!')

    # determine if the sample_sheet_file can be located at the provided path
    if not samplesheet.exists():
        quit_with_error('User requested bcl2fastq2 be performed, but the path provided to the SampleSheet.csv file'
                        'can not be verified.\n Please ensure that the path to the SampleSheet.csv file provided '
                        'using --sample-sheet is correct!')

    # verify that the sample sheet has the proper name that bcl2fastq2 will accept
    if not samplesheet.name == 'SampleSheet.csv':
        quit_with_error('The sample sheet provided by the user has an unacceptable name. bcl2fastq2 requires that the'
                        'sample sheet name be SampleSheet.csv (case sensitive!)')

    # verify that the sample sheet meets the no duplicates criteria for samples and barcode sets that will cause it to
    # fail silently
    sample_sheet_data = {}
    sample_list = []

    with open(str(samplesheet), 'r') as sample_sheet:
        sample_sheet_reader = csv.DictReader(sample_sheet,
                                             fieldnames=['Sample_ID', '', '', '', '', 'index', '', 'index2'],
                                             delimiter=',')
        sample_number = 0
        for row in sample_sheet_reader:
            if len(row['index']) == len(row['index2']) and len(row['index']) == 8:
                concatenated_indices = row['index'] + row['index2']
                sample_number += 1
                if row['Sample_ID'] in sample_sheet_data.keys():
                    quit_with_error('Duplicate sample names detected in the provided sample sheet: ' + row['Sample_ID'])
                elif concatenated_indices in sample_sheet_data.values():
                    quit_with_error('Duplicate barcode combinations detected in the provided sample sheet: I7=' +
                                    row['index'] + ' I5=' + row['index2'])
                else:
                    sample_sheet_data[row['Sample_ID']] = concatenated_indices
                    sample_list.append(row['Sample_ID'])

    # make sure that the SampleSheet.csv file is in the top level Illumina run directory, as that is where bcl2fastq
    # will be looking for it
    if samplesheet.parent == run_directory:
        return sample_list, samplesheet

    # since the SampleSheet.csv is not in the proper directory, we will copy it there
    copied_samplesheet = run_directory / samplesheet.name
    subprocess.run(['cp', str(samplesheet), str(copied_samplesheet)])

    # verify the copy took place
    if not copied_samplesheet.exists():
        quit_with_error('The sample sheet failed to move to the run directory.')

    return sample_list, copied_samplesheet


def query_yes_no(question, default=None):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def round_down(num, divisor):
    return num - (num % divisor)


@contextmanager
def suppress_stdout_stderr():
    """
    A context manager that temporarily directs stdout and stderr to devnull
    """
    with open (os.devnull,'w') as fnull:
        with redirect_stdout(fnull) as out, redirect_stderr(fnull) as err:
            yield (err, out)


def ramdisk_handler(mount=False, unmount=False, ramdisk_directory=None, directory_to_load=None, colors=1,
                    stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None):
    """
    RAMDISK. RAMDISK.

    All hail the SPPPEEEEEEEED.

    This method will attempt to mount or unmount a ramdisk for use in speeding up various aspects of the the selected
    analyses.

    :param mount: if mount is True, we will attempt to mount the ramdisk
    :param unmount: if unmount is True, we will attempt to unmount the ramdisk
    :param ramdisk_directory: if a path is provided, we will verify it exists and use it
    :param directory_to_load:  if a path is provided, we will verify it exists and use it
    :return: True if the requested action was accomplished (mount or unmount), False if not
    """

    if (mount and (not ramdisk_directory or not directory_to_load)) or (unmount and not ramdisk_directory):
        return False

    mlog.log('Checking the current status of the ramdisk\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    # verify that either mount or unmount are true (otherwise return False for impossible request)
    if (mount and unmount) or (not mount and not unmount):
        return False

    # verify that the ramdisk directory exists
    if ramdisk_directory:
        ramdisk_directory = Path(ramdisk_directory).expanduser().resolve()
        if not ramdisk_directory.is_dir():
            mlog.log('Unable to locate the ramdisk directory at the specified path',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            return False

    # verify that the directory_to_load exists
    if directory_to_load:
        directory_to_load = Path(directory_to_load).expanduser().resolve()
        if not directory_to_load.is_dir():
            mlog.log('Unable to locate the database directory at the specified path',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            return False

    # verify whether or not the ramdisk is mounted currently. This is accomplished by running
    # 'df -h | grep [ramisk_directory]' and capturing the stdout from the terminal. If the ramdisk is mounted the length
    # of the stdout and by extension mount_status will be non-zero.
    df_proc = str(subprocess.run(['df', '-h'], stdout=subprocess.PIPE).stdout)
    if str(ramdisk_directory) in df_proc:
        mlog.log('The ramdisk is currently mounted\n',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        mount_status = True
    else:
        mount_status = False

    # if mount was requested, handle the mounting and loading of the ramdisk as necessary
    if mount:

        # mount the ramdisk is it is not currently mounted
        if not mount_status:
            mlog.log('The ramdisk is not currently mounted. Mounting now...',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            # mount the ramdisk
            subprocess.run(['mount', str(ramdisk_directory)])
            # use the mount_status method above to verify successful mounting
            df_proc = str(subprocess.run(['df', '-h'], stdout=subprocess.PIPE).stdout)
            if str(ramdisk_directory) in df_proc:
                mlog.log('The ramdisk has been successfully mounted!\n',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
            else:
                mlog.log('Unable to mount the ramdisk',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
                return False

        # check to see if the database directory is currently mounted
        mlog.log('Checking to see if the database directory has already been loaded on to the ramdisk...',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        mounted_directory = ramdisk_directory / directory_to_load.name
        if mounted_directory.exists():
            # There are two possible scenarios where MIGHT might find that the database is already present on the
            # ramdisk. Possibility #1: Another application has already loaded the required files onto the ramdisk
            # completely. In this case we can just use it. Possibility #2: Another application IS IN THE PROCESS of
            # loading the required files. In this case we will sleep for 30s and then check again to see if the files
            # are fully loaded
            waiting_to_load = True
            while waiting_to_load:
                dcmp = dircmp(str(mounted_directory), str(directory_to_load))
                # Possibility #1: Another application has already loaded the required files onto the ramdisk
                if len(dcmp.right_only) == 0 and len(dcmp.left_only) == 0 and len(dcmp.diff_files) == 0:
                    mlog.log('The ramdisk is mounted, the database is loaded, time to boogie!\n',
                             colors=colors,
                             stdout_verbosity_level=stdout_verbosity_level,
                             log_file_verbosity_level=log_file_verbosity_level,
                             log_file=log_file)
                    return True
                # Possibility  # 2: Another application IS IN THE PROCESS of loading the required files. In this case
                # we will sleep for 30s and then check again to see if the files are fully loaded
                else:
                    mlog.log('The ramdisk is currently being loaded by another application. Waiting...!\n',
                             colors=colors,
                             stdout_verbosity_level=stdout_verbosity_level,
                             log_file_verbosity_level=log_file_verbosity_level,
                             log_file=log_file)
                    time.sleep(30)

        # load the database directory on to the ramdisk
        mlog.log('Database directory has NOT been loaded to the ramdisk.\n',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        mlog.log('Copying the database directory contents to the ramdisk...',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        subprocess.run(['cp', '-r', str(directory_to_load), str(ramdisk_directory)])

        time.sleep(1)

        # verify that the directory was loaded successfully
        dcmp = dircmp(str(mounted_directory), str(directory_to_load))
        if len(dcmp.right_only) > 0:
            mlog.log('The ramdisk failed to load the full complement of files from the database directory',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            return False
        if len(dcmp.diff_files) > 0:
            mlog.log('Discrepancies exist between the files in the database directory and the ramdisk',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            return False

        mlog.log('Database directory successfully loaded to ramdisk. Ready for use!\n',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        return True

    # if unmount was requested, handle the unmounting of the ramdisk
    if unmount:
        if mount_status:
            # umount the ramdisk
            mlog.log('Unmounting the ramdisk...\n',
                     colors=colors,
                     stdout_verbosity_level=stdout_verbosity_level,
                     log_file_verbosity_level=log_file_verbosity_level,
                     log_file=log_file)
            subprocess.run(['umount', str(ramdisk_directory)])

            # verify that the unmount was successful
            df_proc = str(subprocess.run(['df', '-h'], stdout=subprocess.PIPE).stdout)
            if str(ramdisk_directory) in df_proc:
                mlog.log('Unmounting failed!\n',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
            else:
                mlog.log('Unmounting complete!\n',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)
        return True


def ascii_art(logo=None):
    """
    Print out the ascii art logo requested by the user
    :param logo: the name of the logo to be printed
    """

    if logo =='andale':
        print(mlog.bold_green("""

               ___               _              _
              /   \   _ _     __| |    __ _    | |     ___
              | - |  | ' \   / _` |   / _` |   | |    / -_)
              |_|_|  |_||_|  \__,_|   \__,_|  _|_|_   \___|
            _|=====|_|====|_|======|_|=====|_|=====|_|=====|
            "`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'

        """))

    if logo =='might':
        print(mlog.bold_green("""
        
        
            .___  ___.  __    _______  __    __  .__________.
            |   \/   | |  |  /  _____||  |  |  | |          |
            |  \  /  | |  | |  |  __  |  |__|  | `---|  |---`
            |  |\/|  | |  | |  | |_ | |   __   |     |  |     
            |  |  |  | |  | |  |__| | |  |  |  |     |  |     
            |__|  |__| |__|  \______| |__|  |__|     |__|     
    
                
        """))



#!/usr/bin/env python
# encoding: utf-8
import subprocess
from pathlib import Path
import might.might_log as mlog
import time


def database_check(update=False, colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None):
    """
    Check to see if we can see a) the AMRFinderPlus database and b) the ariba reference derived from (a)
    """

    mlog.log_section_header('AMR Database Validation',
                            colors=colors,
                            stdout_verbosity_level=stdout_verbosity_level,
                            log_file_verbosity_level=log_file_verbosity_level,
                            log_file=log_file)
    mlog.log_explanation('In order to perform AMR gene identification, Andale requires that a copy of the NCBI '
                         'AMRFinderPlus database is installed on your machine in an accessible location. Additionally, '
                         'a reference directory prepared using the AMR_CDS file by ARIBA is required to perform AMR '
                         'gene identification from read files.',
                         colors=colors,
                         stdout_verbosity_level=stdout_verbosity_level,
                         log_file_verbosity_level=log_file_verbosity_level,
                         log_file=log_file)

    mlog.log('Checking for the required databases...',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    # by default we will NOT remake the ARIBA reference unless we have to
    refresh = False

    # get the path of the directory where andale.py is installed. This is where the ariba reference will be stored
    conda_bin_directory = Path(__file__).resolve().parent

    # find the path to the amrfinder executable, as it installs its database relative to this location
    proc = subprocess.Popen(['which', 'amrfinder'], stdout=subprocess.PIPE)
    amrfinder_path = Path(str(proc.stdout.read().strip())[2:-1])
    amrfinder_database = amrfinder_path.parent.parent / 'share' / 'amrfinderplus' / 'data' / 'latest'
    amrfinder_database = Path(amrfinder_database).resolve()
    ariba_database = conda_bin_directory / 'andale_ariba_amrfinder_reference'

    # check for a lock file to see if another instance of MIGHT/Andale is attempting to prepare the database. If a lock
    # file is found, enter a while loop that checks periodically if the lock file has been removed, then proceed. If no
    # lock file is encountered, create one until all database preparation methods are complete

    database_lock_file = amrfinder_database.parent / 'andale.lock'

    while database_lock_file.exists():
        mlog.log('encountered database lock file. Waiting for permission to proceed',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)
        time.sleep(10)

    # check to see if the AMRFinderPlus database is present. If the AMRFinderPlus database directory can't be located,
    # call the get_amrfinderplus_database() method to download it

    if not amrfinder_database.is_dir() or update:
        get_amrfinderplus_database(database_lock_file,
                                   colors=colors,
                                   stdout_verbosity_level=stdout_verbosity_level,
                                   log_file_verbosity_level=log_file_verbosity_level,
                                   log_file=log_file)
        # if we are downloading a new AMRFinder database we will also want to remake the ariba reference if it already
        # exists
        if ariba_database.is_dir():
            refresh = True
    else:
        mlog.log('AMRFinderPlus database directory located',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # check to see if ariba prepareref has been run on the AMRFinderPlus database. If the ariba reference directory
    # can't be located or needs to be remade, call the prepare_ariba_reference() method to construct it
    if not ariba_database.is_dir() or refresh:
        prepare_ariba_reference(ariba_database,
                                amrfinder_database,
                                refresh,
                                colors=colors,
                                stdout_verbosity_level=stdout_verbosity_level,
                                log_file_verbosity_level=log_file_verbosity_level,
                                log_file=log_file)
    else:
        mlog.log('ARIBA reference directory located',
                 colors=colors,
                 stdout_verbosity_level=stdout_verbosity_level,
                 log_file_verbosity_level=log_file_verbosity_level,
                 log_file=log_file)

    # remove lock file if it exists
    if database_lock_file.exists():
        subprocess.run(['rm', str(database_lock_file)])

    return amrfinder_database, ariba_database


def get_amrfinderplus_database(database_lock_file, colors=1, stdout_verbosity_level=1, log_file_verbosity_level=1,
                               log_file=None):
    """
    Download the latest version of the AMRFinderPlus database from the ncbi via rsync
    """
    subprocess.run(['touch', str(database_lock_file)])
    mlog.log('Downloading AMRFinderPlus database from NCBI...',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)
    # utilize the AMRFinderPlus built in utility to download/update the amrfinder database
    subprocess.run(['amrfinder', '--update'], stderr=subprocess.STDOUT, stdout=subprocess.DEVNULL)
    mlog.log('Download complete!\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)
    return


def prepare_ariba_reference(ariba_database, amrfinder_database, refresh, colors=1, stdout_verbosity_level=1,
                            log_file_verbosity_level=1, log_file=None):
    """
    Prepare the ariba reference directory using the AMRFinderPlus database
    """
    mlog.log('Preparing the ARIBA reference using the AMRFinderPlus database...\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    # if refresh is true we first need to remove the existing ariba reference directory
    if refresh:
        subprocess.run(['rm', '-r', ariba_database])

    # we require the 'AMR_CDS' file from the database to construct the ariba reference
    amrfinder_CDS_file = amrfinder_database / 'AMR_CDS'
    ariba_prepareref_cmd = 'ariba prepareref --all_coding yes -f ' + str(amrfinder_CDS_file) + ' ' + str(ariba_database)

    mlog.log('ARIBA prepareref command: '+ariba_prepareref_cmd,
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    subprocess.run(ariba_prepareref_cmd, shell=True, stderr=subprocess.STDOUT, stdout=subprocess.DEVNULL)

    mlog.log('ARIBA reference is ready for use!\n',
             colors=colors,
             stdout_verbosity_level=stdout_verbosity_level,
             log_file_verbosity_level=log_file_verbosity_level,
             log_file=log_file)

    return


if __name__ == '__main__':
    database_check()


#!/usr/bin/env python
# encoding: utf-8

import setuptools

with open("README.md", "r") as readme:
    long_description = readme.read()

setuptools.setup(
    name="mrsn-might",
    version="1.0.3",
    author="Brendan Corey",
    author_email="brendan.w.corey.ctr@mail.mil",
    description="MIGHT: MRSN Integrated Genome Handling Tool for bacterial clinical isolates",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/mrsn-bio/might/",
    license="GPLv3",
    packages=setuptools.find_packages(),
    entry_points={'console_scripts': ['might = might.might:premain',
                                      'allmight = might.allmight:main']
                  },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    python_requires='>=3.6'
)


"""
Copyright 2017 Ryan Wick (rrwick@gmail.com)
https://github.com/rrwick/Unicycler
This module contains a Unicycler class for writing output to both stdout and a log file.
This file is part of Unicycler. Unicycler is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Unicycler is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Unicycler. If
not, see <http://www.gnu.org/licenses/>.
"""
import sys
import datetime
import re
import shutil
import textwrap
import subprocess
from pathlib import Path


def initialize_logger_attributes(sample, log_filename=None, stdout_verbosity_level=1, log_file_verbosity_level=None):
    sample.log_filename = log_filename

    # initialize the log file
    if not Path(str(sample.log_filename)).expanduser().resolve().exists():
        subprocess.run(['touch', str(sample.log_filename)])

    # Determine if the terminal supports colours or not.
    try:
        sample.colors = int(subprocess.check_output(['tput', 'colors']).decode().strip())
    except (ValueError, subprocess.CalledProcessError, FileNotFoundError, AttributeError):
        sample.colors = 1

    # There are two verbosity levels: one for stdout and one for the log file. They are the
    # same, except that the log file verbosity level is never 0.
    sample.stdout_verbosity_level = stdout_verbosity_level
    if not log_file_verbosity_level:
        sample.log_file_verbosity_level = stdout_verbosity_level
    else:
        sample.log_file_verbosity_level = log_file_verbosity_level
    sample.log_file_verbosity_level = max(1, sample.log_file_verbosity_level)


def open_log_file(log_file=None):
    if log_file:
        log_file_exists = Path(str(log_file)).expanduser().resolve()
        return open(str(log_file_exists), 'at', 1, encoding='utf8')  # line buffering


def close_log_file(log_file_handle):
    if log_file_handle and not log_file_handle.closed:
        log_file_handle.close()


def log(text, colors=1, verbosity=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None, stderr=False,
        end='\n', print_to_screen=True, write_to_log_file=True):

    text_no_formatting = remove_formatting(text)

    # The text is printed to the screen with ANSI formatting, if supported. If there are only 8
    # colours available, then remove the 'dim' format which doesn't work.
    if stderr or (verbosity <= stdout_verbosity_level and print_to_screen):
        if colors <= 1:
            text = text_no_formatting
        elif colors <= 8:
            text = remove_dim_formatting(text)
        if stderr:
            print(text, file=sys.stderr, end=end, flush=True)
        else:
            print(text, end=end, flush=True)

    # The text is written to file without ANSI formatting.
    if log_file and verbosity <= log_file_verbosity_level and write_to_log_file:

        # open the log file for writing
        log_file_handle = open_log_file(log_file=log_file)

        log_file_handle.write(text_no_formatting)
        log_file_handle.write('\n')

        # close the log file
        close_log_file(log_file_handle)


def log_section_header(message, colors=1, verbosity=1, stdout_verbosity_level=1, log_file_verbosity_level=1,
                       log_file=None, single_newline=False):
    """
    Logs a section header. Also underlines the header using a row of dashes to the log file
    (because log files don't have ANSI formatting).
    """
    if single_newline:
        log('',
            colors=colors,
            verbosity=verbosity,
            stdout_verbosity_level=stdout_verbosity_level,
            log_file_verbosity_level=log_file_verbosity_level,
            log_file=log_file)
    else:
        log('\n',
            colors=colors,
            verbosity=verbosity,
            stdout_verbosity_level=stdout_verbosity_level,
            log_file_verbosity_level=log_file_verbosity_level,
            log_file=log_file)

    time = get_timestamp()
    time_str = '(' + time + ')'
    if colors > 8:
        time_str = dim(time_str)
    log(bold_yellow_underline(message) + ' ' + time_str,
        colors=colors,
        verbosity=verbosity,
        stdout_verbosity_level=stdout_verbosity_level,
        log_file_verbosity_level=log_file_verbosity_level,
        log_file=log_file)
    log('-' * (len(message) + 3 + len(time)),
        colors=colors,
        verbosity=verbosity,
        stdout_verbosity_level=stdout_verbosity_level,
        log_file_verbosity_level=log_file_verbosity_level,
        log_file=log_file,
        print_to_screen=False)


def log_progress_line(sample, completed, total, base_pairs=None, end_newline=False):
    """
    Logs a progress line using a carriage return to overwrite the previous progress line. Only the
    final progress line will be written to the log file.
    """
    progress_str = int_to_str(completed) + ' / ' + int_to_str(total)
    if total > 0:
        percent = 100.0 * completed / total
    else:
        percent = 0.0
    progress_str += ' (' + '%.1f' % percent + '%)'
    if base_pairs is not None:
        progress_str += ' - ' + int_to_str(base_pairs) + ' bp'

    end_char = '\n' if end_newline else ''
    sample.log('\r' + progress_str, end=end_char, write_to_log_file=False)
    if end_newline:
        log(progress_str, print_to_screen=False)


def log_explanation(text, colors=1, verbosity=1, stdout_verbosity_level=1, log_file_verbosity_level=1, log_file=None,
                    print_to_screen=True, write_to_log_file=True, extra_empty_lines_after=1, indent_size=4):
    """
    This function writes explanatory text to the screen. It is wrapped to the terminal width for
    stdout but not wrapped for the log file.
    """
    text = ' ' * indent_size + text
    if print_to_screen:
        terminal_width = shutil.get_terminal_size().columns
        for line in textwrap.wrap(text, width=terminal_width - 1):
            if colors > 8:
                formatted_text = dim(line)
            else:
                formatted_text = line
            log(formatted_text,
                colors=colors,
                verbosity=verbosity,
                stdout_verbosity_level=stdout_verbosity_level,
                log_file_verbosity_level=log_file_verbosity_level,
                log_file=log_file,
                print_to_screen=True,
                write_to_log_file=False)
    if write_to_log_file:
        log(text,
            colors=colors,
            verbosity=verbosity,
            stdout_verbosity_level=stdout_verbosity_level,
            log_file_verbosity_level=log_file_verbosity_level,
            log_file=log_file,
            print_to_screen=False,
            write_to_log_file=True)

    for _ in range(extra_empty_lines_after):
        log('',
            colors=colors,
            verbosity=verbosity,
            stdout_verbosity_level=stdout_verbosity_level,
            log_file_verbosity_level=log_file_verbosity_level,
            log_file=log_file,
            print_to_screen=print_to_screen,
            write_to_log_file=write_to_log_file)


def log_number_list(sample, numbers, verbosity=1, print_to_screen=True, write_to_log_file=True,
                    indent_size=4):
    """
    Some lists of numbers are long, so this function makes them wrap nicely when displayed on
    stdout.
    """
    text = ' ' * indent_size + ', '.join(str(x) for x in numbers)
    if print_to_screen:
        terminal_width = shutil.get_terminal_size().columns
        for line in textwrap.wrap(text, width=terminal_width - 1):
            sample.log(line, verbosity=verbosity, print_to_screen=True, write_to_log_file=False)
    if write_to_log_file:
        sample.log(text, verbosity=verbosity, print_to_screen=False, write_to_log_file=True)


def int_to_str(num, max_num=0):
    if num is None:
        num_str = 'n/a'
    else:
        num_str = '{:,}'.format(num)
    max_str = '{:,}'.format(int(max_num))
    return num_str.rjust(len(max_str))


def get_timestamp():
    return '{:%Y-%m-%d %H:%M:%S}'.format(datetime.datetime.now())


END_FORMATTING = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'
YELLOW = '\033[93m'
DIM = '\033[2m'
RED = '\033[31m'
GREEN = '\033[32m'


def green(text):
    return GREEN + text + END_FORMATTING


def bold_green(text):
    return GREEN + BOLD + text + END_FORMATTING


def bold_red(text):
    return RED + BOLD + text + END_FORMATTING


def bold(text):
    return BOLD + text + END_FORMATTING


def bold_yellow_underline(text):
    return YELLOW + BOLD + UNDERLINE + text + END_FORMATTING


def dim(text):
    return DIM + text + END_FORMATTING


def remove_formatting(text):
    return re.sub('\033.*?m', '', text)


def remove_dim_formatting(text):
    return re.sub('\033\[2m', '', text)